; (
; Gibt selektierte Entitaeten in CNC-Format aus
;     Polylines, die Kreisboegen (arc) werden durch Linien angenaehert.
;
(defun CNCError (st)
   (if (/= st "Function cancelled")
      (princ (strcat "\nCNC-Error: " st))
   )
   (ModeReset)
   (close FileOutDebug)
   (close FileOutCNC)
   (setq *error* OldErr)
   (princ)
)

(defun ColorNumber (s / i col)
   (setq i 1)
   (setq col "")
   (setq s (strcat s " "))


   (while ( /= (substr s i 1) " ")
       (setq col (strcat col (substr s i 1)) )
       (setq i (1+ i) )
   )
   (setq col col)
)


(defun ModeSet ()
   (setq scmdecho (getvar "cmdecho" ) )
   (setq slayer (getvar "clayer") )
   (setq scolor (getvar "cecolor")  )
   (setq scolor (ColorNumber scolor) )
   (setvar "cmdecho" 0)
   (command "Layer" "Make" CNCLayerName "")
   (command "Color" CNCColorName )
)
;----- Mode Reset -- Resets previously saves system variables ---

(defun ModeReset ()
   (command "LAYER" "Make" slayer "")
   (command "Color" scolor)
   (setvar "cmdecho" scmdecho)
)

;--------- Convert bulge information -------------------
; AutoLISP function to convert from Polyline "Bulge" representation
; of an arc to AutoCAD's normal "center, radius, start/end angles"
; form of arc.  This function applies the bulge between two adjacent
; vertices.  It assumes that global symbols "sp", "ep", and "bulge"
; contain the current vertex (start point), next vertex (end point),
; and bulge, respectively.
; Its makes a list with:
;   (center radius st-ang  end-ang clockwise)

(defun ConvBulge (sp ep bulge / x1 x2 y1 y2 cotbce
                               center radius st-ang end-ang clockwise)
  (setq x1 (car  sp) x2 (car  ep))
  (setq y1 (cadr sp) y2 (cadr ep))
  (setq cotbce (/ (- (/ 1.0 bulge) bulge) 2.0))

  ; Compute center point and radius

  (setq center (list (/ (+ x1 x2 (- (* (- y2 y1) cotbce))) 2.0)
                     (/ (+ y1 y2    (* (- x2 x1) cotbce) ) 2.0)
                     (caddr sp)
               )
  )
  (setq radius (distance center sp))

  ; Compute start and end angles

  (setq st-ang  (atan (- y1 (cadr center)) (- x1 (car center))))
  (setq end-ang  (atan (- y2 (cadr center)) (- x2 (car center))))
  (if (< st-ang 0.0)                      ; Eliminate negative angles
     (setq st-ang (+ st-ang (* 2.0 pi)))
  )
  (if (< end-ang 0.0)
     (setq end-ang (+ end-ang (* 2.0 pi)))
  )

  (setq clockwise (< bulge 0.0) )

  (if (< bulge 0.0)                   ; Swap angles if clockwise
     (progn
        (setq temp st-ang)
        (setq st-ang end-ang)
        (setq end-ang temp)
     )
  )
  (list center radius st-ang end-ang clockwise)
) ; ConvBulge

;----- Converts radians to degrees ------------------------------

(defun rtd (r)
 (* 180 (/ r pi))
)


(defun Header_1 ()
   (print_cnc "; PROJEKTNUMMER EINGEBEN")
   (print_cnc "")
   (print_cnc "G64")
   (print_cnc "G91 Z-0.1")
   (print_cnc "G90")
   (print_cnc "")
   (print_cnc "")
   
)


(defun Header_2 ()
   (print_cnc "")
   (print_cnc "%NR*1  (-D11-000000- - -TEXT)")
)


(defun WriteHeader (NR)
   (cond
       ( (= NR 1)
             (Header_1) )
       ( (= NR 2)
             (Header_2) )
       ( T
              'Keine_Maschine_definiert)
    ) ; cond
) ; WriteHeader

(defun WriteTail ()
   (print_cnc "M17")
   )

(defun print_list (el)
   (repeat (length el)
      (princ (car el) FileOutDebug)
      (princ "\n" FileOutDebug)
      (setq el (cdr el))
   )
)

(defun print_cnc (x)
   (princ " ----- cnc ------: " FileOutDebug)
   (princ x FileOutDebug)
   (princ "\n" FileOutDebug)

   (princ x FileOutCNC)
   (princ "\n" FileOutCNC)
)

(defun MakeZeros (n / s1)
    (setq s1 "")
    (if (and ( >= n 0 ) ( <= n 9 ))
       (setq s1 "000" )
    )
    (if (and ( >= n 10 ) ( <= n 99 ))
       (setq s1 "00"  )
    )
    (if (and ( >= n 100 ) ( <= n 999 ))
       (setq s1 "0" )
    )
    (strcat s1 (rtos n 2 0) )
)

(defun MakeG1String (n k / s1 s2 x y z lx ly lz)
    (setq   x   (car k)
            y   (cadr k)
            z   (caddr k)
            lx  (car LastKoord)
            ly  (cadr LastKoord)
            lz  (caddr LastKoord) )

    (setq s1 (strcat "N" (MakeZeros n) " G1") )
    (setq s2 "")
    (if (/= lx x )
       (progn
          (setq s2 (strcat s2 " X" (rtos (- x (CAR transkoord))  2 KommaStellen) ) )
          (setq lx x)
       )
    ) ; if
    (if (/= ly y )
       (progn
          (setq s2 (strcat s2 " Y" (rtos (- y (CADR TRANSKOORD)) 2 KommaStellen) ) )
          (setq ly y)
       )
    ) ; if
    (if (/= lz z )
       (progn
          (setq s2 (strcat s2 " Z" (rtos (- z (caddr transkoord))  2 KommaStellen) ) )
          (setq lz z)
       )
    ) ; if
   (setq LastKoord (list x y z) )
   (if (/= s2 "")
      (setq s1 (strcat s1 s2))
      ;else
      (setq s1 "")
   )
)

(defun MakeBulgeString (n p1 p2 bul / s1 cb center x y z)
    (setq s1 (strcat "N" (MakeZeros n) ) )
    (setq cb (ConvBulge p1 p2 bul))
    (setq center (car cb))
    (if (last cb)
       (setq s1 (strcat s1 " G2 ") )
       ;else
       (setq s1 (strcat s1 " G3 ") )
    ) ; if


 (setq s1 (strcat s1 "X" (rtos (- (car p1) (car Transkoord)) 2 KommaStellen)
                    " Y" (rtos (- (cadr p1) (cadr Transkoord)) 2 KommaStellen)
                " I" (rtos (- (car p1) (car center)) 2 KommaStellen)
                " J" (rtos (- (cadr p1) (cadr center)) 2 KommaStellen)
             )
    )
    (setq x (car p1)  )
    (setq y (cadr p1) )
    (setq z (caddr p1) )
    (setq LastKoord (list x y z) )

    (setq s1 s1)
)


(defun AnaVertex (el / name koord arc-info bv)
   (setq name (cdr (assoc 0 el)) )
   (if (= name "VERTEX")
      (progn
         (setq koord (cdr (assoc 10 el)) )
         (setq bv (cdr (assoc 42 el)) )
         (if (= p1 nil)
             (setq p1 koord)
             ;else
             (progn
                (setq p2 p1)
                (setq p1 koord)
             ) ; progn
         ) ; if

         (if Bulge
            (setq str (MakeBulgeString (1+ LineAnz) p1 p2 BulgeVal))
            ;else
            (setq str (MakeG1String (1+ LineAnz) koord) )
         ) ; if Bulge
         (if (/= str "")
            (progn
                (setq LineAnz (1+ LineAnz) )
                (print_cnc str)
            )
        )
        (if (/= bv 0)
           (progn
              (setq BulgeVal bv)
              (setq Bulge T)
           )
           ;else
           (progn
              (setq BulgeVal 0)
              (setq Bulge nil)
           )
        ); if

      ) ; prog
   ); if
) ; AnaVertex

(defun AnaPolyline (i enam / elist e1 name Bulge BulgeVal p1 p2)
    (setq p1 nil
          p2 nil)
    (setq Bulge nil)
    (setq BulgeVal 0)
    (setq elist (entget enam))
    (princ (+ i 1) FileOutDebug)
    (princ ". Entity: \n" FileOutDebug )
    (print_list elist)
    ; Unterentitaet bis SEQEND analysieren
    (setq e1 (entnext enam))
    (while (AND e1 (/= (cdr (assoc 0 (entget e1) ) ) "SEQEND") )
         (print_list (entget e1) )
         (AnaVertex (entget e1) )
         (setq e1 (entnext e1) )
     ) ; while
     (princ "\n\n " FileOutDebug)
) ; AnaPolyline

(defun EqualKoord (p1 p2 / e)
;   (print "Equal:")
;   (print p1)
;   (print p2)
   (setq e   (AND (= (car p1) (car p2))
                  (= (cadr p1) (cadr p2))
                  (= (caddr p1) (caddr p2))
             )
   )
;   (print e)
   (setq e e)
)

(defun LastKoordError (p1 p2 p3 p4)
   (princ "\nAnfangspunkt:\n")
   (print p1)
   (print p2)
   (print p3)
   (print p4)
   (princ "\n")
)

(defun AnaLine (i enam / elist p1 p2 p Fehler)
    (setq p1 nil
          p2 nil
          Fehler 0)
    (setq elist (entget enam))
    (princ (+ i 1) FileOutDebug)
    (princ ". Entity: \n" FileOutDebug )
    (print_list elist)
    (setq p1 (cdr (assoc 10 elist)) )
    (setq p2 (cdr (assoc 11 elist)) )
    (setq p LastKoord)
    (if (EqualKoord p1 LastKoord)
       (setq p p2)
       ;else
       (if (EqualKoord p2 LastKoord)
          (setq p p1)
          ;else
          (progn
             (setq Fehler 1)
             (LastKoordError p1 p2 LastKoord)
          ) ;progn
       ) ; if
    ); if
    (if (= Fehler 0)
       (progn
          (setq str (MakeG1String (1+ LineAnz) p) )
          (if (/= str "")
             (progn
                (setq LineAnz (1+ LineAnz) )
                (print_cnc str)
             )
          )
       ) ;progn
    ) ; if
) ; AnaLine

(defun AnaArc (i enam / elist p1 p2)
    (setq p1 nil
          p2 nil)
    (setq elist (entget enam))
    (princ (+ i 1) FileOutDebug)
    (princ ". Entity: \n" FileOutDebug )
    (print_list elist)
    (setq p1 (cdr (assoc 10 elist)) )
    (setq p2 (cdr (assoc 11 elist)) )

    (setq str (MakeG1String (1+ LineAnz) p1) )
    (if (/= str "")
       (progn
          (setq LineAnz (1+ LineAnz) )
          (print_cnc str)
       )
    )
    (setq str (MakeG1String (1+ LineAnz) p2) )
    (if (/= str "")
       (progn
          (setq LineAnz (1+ LineAnz) )
          (print_cnc str)
       )
    )
) ; AnaLine


(defun InitVariables (/ ret)
; Falls ein Fehler auftritt: nil
    (setq ret T )

    (setq FileName "" )
    (while (= FileName "")
        (setq FileName (getstring "Dateiname: "))
    )

    (setq MaschinenNr 0)
    (while (OR (< MaschinenNr 1) (> MaschinenNr 2) )
       (setq MaschinenNr (getInt "Welche Maschine (1 (emco-mill-155) od.2(FP3NC %*)):") )
    )

    (initget 16)
    (setq TransKoord (getPoint "Verschiebungs-Punkt eingeben (x y z): ") )

    (setq FileOutDebug (open (strcat FileName ".deb") "w"))
    (setq FileOutCNC   (open (strcat FileName ".spf") "w"))
    (setq ret (AND (/= FileOutDebug nil) (/= FileOutCNC nil) ) )
    (if ret
       ()
       (princ "\nFehler beim File-Oeffnen\n")
   )
   (setq ret ret)
 ); InitVariables


;
; *************************************************
; *
; * Hauptprogramm:
; *
; *************************************************

(defun c:cnc (/ OldErr KommaStellen LineAnz Auswahl AuswahlAnz FileName
                FileOutDebug  FileOutCNC i EName LastKoord MaschinenNr
                TransKoord CNCLayerName CNCLayerColor slayer scolor Genauigkeit)
   (setq KommaStellen   3
         Genauigkeit    4
         LineAnz        0
         Auswahl        nil
         AuswahlAnz     0
         FileName       ""
         EntNr          0
         EName          nil
         LastKoord      (list -999999.9999 -999999.9999 -999999.9999)
         TransKoord     (list 0.0 0.0 0.0)
         MaschinenNr    0
         CNCLayerName   "CNC"
         CNCColorName   "11"
         slayer         0
         scolor         0
         scmdecho       0
         olderr         *error*
         *error*        *error* )

   (ModeSet) ; alte Werte sichern, neue setzen
   (c:cnc_on)
   ; Auswahl der Entitaeten:
   (setq Auswahl (ssget ))
   (if Auswahl
      (setq AuswahlAnz (sslength Auswahl) )
   )
   (if (AND (> AuswahlAnz 0) (InitVariables) )
     (progn
         (WriteHeader MaschinenNr)
         (setq EntNr 0)
         (repeat AuswahlAnz
            (setq EName (ssname Auswahl EntNr))
            (setq ElementName (cdr (assoc 0 (entget EName)) ))
            (cond
                ( (= ElementName "POLYLINE")
                      (AnaPolyLine EntNr EName) )
                ( (= ElementName "LINE")
                      (AnaLine EntNr EName) )
                ( (= ElemntName "ARC")
                      (AnaArc EntNr EName) )
                ( T
                    'Nicht_implementiert)
            ) ; cond
            (setq EntNr (+ EntNr 1))
         ) ; repeat
         (WriteTail)
         (close FileOutDebug)
         (close FileOutCNC)
         (princ "\nFertig ...\n")
      ) ; progn
      ;else
      (progn
         (princ "\nSie sollten schon etwas auswaehlen ...\n")
      ) ; progn

   ) ; if
   (ModeReset)
   (setq *error* olderr)
   (c:cnc_off)
   (princ )
) ; c:cnc


(defun c:cnc_copy (/ OldErr Auswahl AuswahlAnz i EName EntNR
                     CNCLayerName CNCLayerColor slayer scolor olderr)
   (setq
         Auswahl        nil
         AuswahlAnz     0
         CNCLayerName   "CNC"
         CNCColorName   "11"
         slayer         0
         scolor         0
         scmdecho       0
         olderr         *error*
         *error*        *error* )

   (ModeSet) ; alte Werte sichern, neue setzen

   ; Auswahl der Entitaeten:
   (setq Auswahl (ssget ))
   (if Auswahl
      (setq AuswahlAnz (sslength Auswahl) )
   )
   (if (> AuswahlAnz 0)
     (progn
         (command "copy" "prev" "" "0,0,0" "")
         (command "change" "prev" "" "prop" "layer" CNCLayerName
                  "color" CNCColorName "")
         (princ "\nDie selektierten Entitaeten wurden kopiert ...\n")
      ) ; progn
      ;else
      (progn
         (princ "\nSie sollten schon etwas auswaehlen ...\n")
      ) ; progn
   ) ; if
   (ModeReset)
   (setq *error* olderr)
   (princ )
) ; c:cnc

(defun c:cnc_on ()
   (command "layer" "off" "*" "y" "on" "cnc" "")
   (princ "")
)

(defun c:cnc_off ()
   (command "layer" "off" "*" "y" "on" "0" "")
   (princ "")
)


; )
